#pragma warning(disable:4800)

#include "Program.h"

#include <cassert>
#include <cstdio>

#include "util/ParameterParser.h"
#include "util/TimeChecker.h"
#include "util/MemoryChecker.h"

#include "fx/SimpleEffects.h"
#include "fx/ToonEffects.h"

#include "flake/build/SnowFlakeBuilder.h"
#include "flake/SpikeSystemRenderer.h"
#include "flake/SpikeRenderers.h"

#include "bmp/BitmapSerializer.h"




CProgram::CProgram()
:	mColorCanvas(0),
	mDepthCanvas(0),
	mDrawing(0),
	mTimeChecker(0),
	mMemoryChecker(0),
	mParams()
{
}

CProgram::~CProgram()
{
	if( mTimeChecker )
		delete mTimeChecker;
	if( mMemoryChecker )
		delete mMemoryChecker;
	if( mColorCanvas )
		delete mColorCanvas;
	if( mDepthCanvas )
		delete mDepthCanvas;
	if( mDrawing )
		delete mDrawing;
}

bool CProgram::processParams( int argc, char** argv )
{
	CParameterParser parser( argc, argv );

	int t;
	bool okTime = parser.parseInt( "-time", mParams.mAllowedTime );
	bool okMem = parser.parseInt( "-mem", mParams.mAllowedMemory );
	bool okCpu = parser.parseInt( "-cpu", mParams.mCPUSpeed );
	
	if( !okTime || !okMem || !okCpu ) {
		printHelp();
		return false;
	}

	// construct time checker
	// reserve some time just to be sure :)
	mTimeChecker = new CTimeChecker( (float)mParams.mAllowedTime );
	mTimeChecker->reserve( (float)mParams.mAllowedTime * 0.5f + 5.0f );

	// memory checker
	// reserve some for usual reasons
	mMemoryChecker = new CMemoryChecker();
	mMemoryChecker->reserve( 4096 );

	// dark
	bool okDark = parser.parseInt( "-dark", t );
	if( !okDark )
		t = 0;
	mParams.mDark = (bool)t;

	// random seed
	bool okSeed = parser.parseInt( "-seed", mParams.mRandomSeed );
	if( !okSeed )
		mParams.mRandomSeed = 0;

	// pre fx
	bool okPreFx = parser.parseInt( "-prefx", t );
	if( !okPreFx )
		t = 1;
	switch( t ) {
	case 0: mParams.mPreEffect = new CUniformFillEffect( mParams.mDark ? CCanvas::rgb(0,0,0) : CCanvas::rgb(255,255,255) ); break;
	case 1: mParams.mPreEffect = new CBackgroundEffect( mParams.mRandomSeed, mParams.mDark ); break;
	case 2: mParams.mPreEffect = new CBackgroundEffect2( mParams.mRandomSeed, mParams.mDark ); break;
	default: mParams.mPreEffect = new CBackgroundEffect( mParams.mRandomSeed, mParams.mDark ); break;
	};

	// post fx
	bool okPostFx = parser.parseInt( "-postfx", t );
	if( !okPostFx )
		t = 0;
	switch( t ) {
	case 0: mParams.mPostEffect = 0; break;
	case 1: mParams.mPostEffect = new CShadeEffect( CCanvas::rgb( 40, 40, 60 ), 0.04f, 0.8f ); break;
	case 2: mParams.mPostEffect = new CShadeEffect( CCanvas::rgb( 40, 40, 60 ), -0.1f, 0.1f ); break;
	case 3: mParams.mPostEffect = new CBlurEffect(1); break;
	default: mParams.mPostEffect = 0; break;
	};

	// spikes
	int spikes;
	bool okSpikes = parser.parseInt( "-spikes", spikes );
	if( !okSpikes )
		spikes = 64;

	// builder
	bool okType = parser.parseInt( "-type", t );
	if( !okType )
		t = 0;
	switch( t ) {
	case 0: mParams.mSystemBuilder = new CSnowFlakeBuilder(spikes); break;
	case 1: mParams.mSystemBuilder = new CSnowFlakeBuilder2(spikes); break;
	case 2: mParams.mSystemBuilder = new CSnowFlakeBuilder3(spikes); break;
	case 3: mParams.mSystemBuilder = new CSnowFlakeBuilder4(spikes); break;
	case 4: mParams.mSystemBuilder = new CXmasTreeBuilder(); break;
	default: mParams.mSystemBuilder = new CSnowFlakeBuilder(spikes); break;
	};

	// iterations
	bool okIter = parser.parseInt( "-iter", mParams.mIterations );
	if( !okIter )
		mParams.mIterations = -1;

	// reserve approx same time&mem for shutdown
	mTimeChecker->reserveElapsed();
	mMemoryChecker->reserveConsumed();

	return true;
}

void CProgram::printHelp() const
{
	printf( "Stuff-and-fluff-generator by Aras Pranckevicius\n" );
	printf( "\n" );
	printf( "Usage: <program-exe> -param value -param value ...\n" );
	printf( "Parameters:\n" );
	printf( "(general - required)\n" );
	printf( "  -time      Allowed time to run (in seconds)\n" );
	printf( "  -mem       dummy\n" );
	printf( "  -cpu       dummy\n" );
	printf( "(tweaks - optional)\n" );
	printf( "  -seed      Seed for random\n" );
	printf( "  -iter      # of iterations (0 to 8 - more at your risk!) overrides -time\n" );
	printf( "  -dark      Should background be dark (0 or 1)\n" );
	printf( "  -prefx     Background fx (0,1,2)\n" );
	printf( "  -postfx    Post fx (0,1,2 or 3)\n" );
	printf( "  -spikes    Initial system size (1 to 100)\n" );
	printf( "  -type      System type (0,1,2,3,4)\n" );
	printf( "\n" );
	printf( "Exaple usage: <program-exe> -time 100 -mem 32 -cpu 600 -dark 1\n" );
}


void CProgram::createCanvases()
{
	printf( "  create bitmaps..." ); fflush( stdout );
	mDrawing = new CDrawing();
	mColorCanvas = new CCanvas( IMG_WIDTH, IMG_HEIGHT );
	mDepthCanvas = new CCanvas( IMG_WIDTH, IMG_HEIGHT );
	mDrawing->prepareCanvas( *mColorCanvas );
	mTextCanvas = CBitmapSerializer<CBuilderBMP>::deserialize( "font.bmp" );
	mTextEffect = new CTextEffect( *mTextCanvas, FONT_WIDTH );
	CTextEffect::initialize();
	printf( " OK\n" );
	
	if( mParams.mPreEffect ) {
		printf( "  pre effect..." ); fflush( stdout );
		mParams.mPreEffect->perform( *mDrawing, *mColorCanvas, *mDepthCanvas );
		printf( " OK\n" );
	}

	// reserve approx same time for shutdown and post effect
	mTimeChecker->reserveElapsed();
}


void CProgram::run()
{
	assert( mParams.mSystemBuilder );

	// create
	printf( "  create system..." ); fflush( stdout );
	CSpikeSystem& ls = CSpikeSystem::getInstance();
	ls.getRandom().seed( mParams.mRandomSeed );
	mParams.mSystemBuilder->buildRules( ls );
	mParams.mSystemBuilder->buildInitial( ls );
	printf( " OK\n" );

	// again, reserve for shutdown
	mTimeChecker->reserveElapsed();
	mMemoryChecker->consumed(); // checkpoint

	// evolve
	printf( "  evolve... " ); fflush( stdout );
	if( mParams.mIterations >= 0 ) {

		mMemoryChecker->consumed();

		if( mParams.mSystemBuilder->getAllowedIterations() < mParams.mIterations )
			mParams.mIterations = mParams.mSystemBuilder->getAllowedIterations();

		for( int i = 0; i < mParams.mIterations; i++ ) {
			// check memory
			int consumed = mMemoryChecker->consumed();
			mMemoryChecker->reserve( consumed * 4 ); // iteration forward
			if( mMemoryChecker->isEnding() ) {
				printf( " (looks like RAM is ending)" );
				break;
			}
			if( ls.getSystem().getSize() > ls.getSystem().MAX_SIZE ) {
				printf( " (detail level reached sub-pixel - there's no point to continue)" );
				break;
			}
			ls.getSystem().evolve(); printf( "%i ", i+1 ); fflush( stdout );
		}
	} else {
		restrictedRun( ls );
	}
	printf( " OK\n" );

	// render - HACK - no time check, evolve should have reserved enough time...
	printf( "  render..." ); fflush( stdout );
	CFuzzyLineRenderer renderer( mParams.mDark );
	CSpikeSystemRenderer<CFuzzyLineRenderer>::render( renderer, *mDrawing, *mColorCanvas, *mDepthCanvas, ls );
	printf( " OK\n" );

	// post fx
	if( mParams.mPostEffect ) {
		printf( "  post effect..." ); fflush( stdout );
		mParams.mPostEffect->perform( *mDrawing, *mColorCanvas, *mDepthCanvas );
		printf( " OK\n" );
	}

	// print size
	char buf[200];
	sprintf( buf, "%i particles", ls.getSystem().getSize() );
	mTextEffect->setText( buf, 5, mColorCanvas->getHeight()-mTextCanvas->getHeight()*2-10, mParams.mDark );
	mTextEffect->perform( *mDrawing, *mColorCanvas, *mDepthCanvas );
	mTextEffect->setText( "size doesnt change, its 1.0 all the time", 25, mColorCanvas->getHeight()-mTextCanvas->getHeight()-5, mParams.mDark );
	mTextEffect->perform( *mDrawing, *mColorCanvas, *mDepthCanvas );
}



void CProgram::restrictedRun( CSpikeSystem& ls )
{
	int iter = 1;

	// kind of HACK-y reserving of time :)
	while( !mTimeChecker->isDeadline() ) {
		int consumed = mMemoryChecker->consumed();
		// looking approx one iteration forward
		mMemoryChecker->reserve( consumed * 4 );
		
		if( mMemoryChecker->isEnding() ) {
			printf( " (looks like RAM is ending)" );
			break;
		}

		double elapsed = mTimeChecker->elapsed();
		// looking approx one iteration forward plus time to render :)
		mTimeChecker->reserve( elapsed * 16.0 );

		if( mTimeChecker->isDeadline() )
			break;
		
		if( ls.getSystem().getSize() > ls.getSystem().MAX_SIZE ) {
			printf( " (detail level reached sub-pixel - there's no point to continue)" );
			break;
		}

		ls.getSystem().evolve(); printf( "%i", iter ); fflush( stdout );
		iter++;
	}
}



void CProgram::finish()
{
	mTextEffect->setText( "stuff n fluff by nearaz", 5, 5, mParams.mDark );
	mTextEffect->perform( *mDrawing, *mColorCanvas, *mDepthCanvas );

	printf( "  saving..." ); fflush( stdout );
	CBitmapSerializer<CBuilderBMP>::serialize( "testSystem.bmp", *mColorCanvas );
	printf( " OK\n" );
}

